@@ -56,18 +56,33 @@ def main():
     "--colors",
     type=str,
     required=False,
-    help="Optional color dictionary. Format is e.g. 'YRI:blue,CEU:green'",
+    help=(
+        "Optional color dictionary. Input can be from the matplotlib list of colors "
+        "or in hexcode. Format is e.g. 'YRI:blue,CEU:green'"
+    ),
 )
-def karyogram(bp, sample, out, title, centromeres, colors):
+@click.option(
+    "-v",
+    "--verbosity",
+    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]),
+    default="INFO",
+    show_default=True,
+    help="The level of verbosity desired",
+)
+def karyogram(bp, sample, out, title, centromeres, colors, verbosity):
     """
     Visualize a karyogram of local ancestry tracks
     """
     from .karyogram import PlotKaryogram
+    from .logging import getLogger
+
+    log = getLogger(name="karyogram", level=verbosity)
 
     if colors is not None:
         colors = dict([item.split(":") for item in colors.split(",")])
+    log.info("Generating Karyogram...")
     PlotKaryogram(
-        bp, sample, out, centromeres_file=centromeres, title=title, colors=colors
+        bp, sample, out, log, centromeres_file=centromeres, title=title, colors=colors
     )
 
 
@@ -76,7 +91,10 @@ def karyogram(bp, sample, out, title, centromeres, colors):
     "--model",
     type=str,
     required=True,
-    help="Admixture model in .dat format. See docs for info.",
+    help=(
+        "Admixture model in .dat format. See File Formats under simgenotype in the "
+        "docs for complete info."
+    ),
 )
 @click.option(
     "--mapdir",
@@ -91,7 +109,11 @@ def karyogram(bp, sample, out, title, centromeres, colors):
     "--out",
     type=str,
     required=True,
-    help="Prefix to name output files.",
+    help=(
+        "Path to desired output file. E.g. /path/to/output.vcf.gz "
+        "Possible outputs are vcf|bcf|vcf.gz|pgen and there will be an "
+        "additional breakpoints output with extension bp e.g. /path/to/output.bp."
+    ),
 )
 @click.option(
     "--chroms",
@@ -116,19 +138,19 @@ def karyogram(bp, sample, out, title, centromeres, colors):
     help="Number of samples to simulate each generation",
 )
 @click.option(
-    "--invcf",
+    "--ref_vcf",
     required=True,
     help=(
-        "VCF file used as reference for creation of simulated samples respective "
-        "genotypes."
+        "VCF or PGEN file used as reference for creation of simulated samples"
+        " respective genotypes."
     ),
 )
 @click.option(
     "--sample_info",
     required=True,
     help=(
-        "File that maps samples from the reference VCF (--invcf) to population codes "
-        "describing the populations in the header of the model file."
+        "File that maps samples from the reference VCF (--invcf) to population"
+        " codes describing the populations in the header of the model file."
     ),
 )
 @click.option(
@@ -136,32 +158,49 @@ def karyogram(bp, sample, out, title, centromeres, colors):
     required=False,
     default=None,
     help=(
-        "Subset the simulation to a specific region in a chromosome using the form"
-        " chrom:start-end. Example 2:1000-2000"
+        "Subset the simulation to a specific region in a chromosome using the"
+        " form chrom:start-end. Example 2:1000-2000"
+    ),
+)
+@click.option(
+    "--pop_field",
+    required=False,
+    is_flag=True,
+    default=False,
+    help=(
+        "Flag for outputting the population field in your VCF output. NOTE this"
+        " flag does not work when your output file is in PGEN format."
+    ),
+)
+@click.option(
+    "--sample_field",
+    required=False,
+    is_flag=True,
+    default=False,
+    help=(
+        "Flag for outputting the sample field in your VCF output. NOTE this"
+        " flag does not work when your output file is in PGEN format."
     ),
 )
 @click.option(
     "--only_breakpoint",
-    hidden=True,
     is_flag=True,
     required=False,
     help=(
-        "Flag used to determine whether to only output breakpoints or continue to "
-        "simulate a vcf file."
+        "Flag used to determine whether to only output breakpoints or"
+        " continue to simulate a vcf file."
     ),
 )
 @click.option(
-    "--verbose",
-    hidden=True,
-    is_flag=True,
-    required=False,
-    help=(
-        "Output time metrics for each section, breakpoint simulation, vcf creation, "
-        "and total exection."
-    ),
+    "-v",
+    "--verbosity",
+    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]),
+    default="INFO",
+    show_default=True,
+    help="The level of verbosity desired",
 )
 def simgenotype(
-    invcf,
+    ref_vcf,
     sample_info,
     model,
     mapdir,
@@ -170,8 +209,10 @@ def simgenotype(
     seed,
     chroms,
     region,
+    pop_field,
+    sample_field,
     only_breakpoint,
-    verbose,
+    verbosity,
 ):
     """
     Simulate admixed genomes under a pre-defined model.
@@ -184,9 +225,16 @@ def simgenotype(
         validate_params,
         write_breakpoints,
     )
+    from .logging import getLogger
 
+    log = getLogger(name="simgenotype", level=verbosity)
     start = time.time()
 
+    # immediately set pop_filed and sample_field flags to false if pgen file
+    if out.endswith(".pgen"):
+        pop_field = False
+        sample_field = False
+
     # parse region and chroms parameters
     if not (chroms or region):
         raise Exception("Either chroms or region must be specified.")
@@ -211,25 +259,39 @@ def simgenotype(
     if mapdir[-1] == "/":
         mapdir = mapdir[:-1]
 
+    # grab prefix from --out for outputting breakpoint
+    out_prefix = re.split(r"(\.vcf|\.bcf|\.vcf\.gz|\.pgen)$", out)[0]
+
     # simulate breakpoints
     popsize = validate_params(
-        model, mapdir, chroms, popsize, invcf, sample_info, region, only_breakpoint
+        model, mapdir, chroms, popsize, ref_vcf, sample_info, region, only_breakpoint
     )
-    samples, breakpoints = simulate_gt(model, mapdir, chroms, region, popsize, seed)
-    breakpoints = write_breakpoints(samples, breakpoints, out)
+    samples, pop_dict, breakpoints = simulate_gt(
+        model, mapdir, chroms, region, popsize, log, seed
+    )
+    breakpoints = write_breakpoints(samples, pop_dict, breakpoints, out_prefix, log)
     bp_end = time.time()
 
     # simulate vcfs
     vcf_start = time.time()
     if not only_breakpoint:
-        # TODO add region functionality
-        output_vcf(breakpoints, chroms, model, invcf, sample_info, region, out)
+        output_vcf(
+            breakpoints,
+            chroms,
+            model,
+            ref_vcf,
+            sample_info,
+            region,
+            pop_field,
+            sample_field,
+            out,
+            log,
+        )
     end = time.time()
 
-    if verbose:
-        print(f"Time elapsed for breakpoint simulation: {bp_end - start}")
-        print(f"Time elapse for creating vcf: {end - vcf_start}")
-        print(f"Time elapsed for simgenotype execution: {end - start}")
+    log.debug(f"Time elapsed for breakpoint simulation: {bp_end - start}")
+    log.debug(f"Time elapsed for creating vcf: {end - vcf_start}")
+    log.debug(f"Total time elapsed for simgenotype execution: {end - start}")
 
 
 @main.command()
@@ -258,6 +320,13 @@ def simgenotype(
     show_default="quantitative trait",
     help="Disease prevalence if simulating a case-control trait",
 )
+@click.option(
+    "--normalize/--no-normalize",
+    is_flag=True,
+    default=True,
+    show_default=True,
+    help="Whether to normalize the genotypes before using them for simulation",
+)
 @click.option(
     "--region",
     type=str,
@@ -322,6 +391,13 @@ def simgenotype(
     show_default="all variants",
     help="If using a PGEN file, read genotypes in chunks of X variants; reduces memory",
 )
+@click.option(
+    "--seed",
+    type=int,
+    default=None,
+    show_default="chosen randomly",
+    help="Use this option across executions to make the output reproducible",
+)
 @click.option(
     "-o",
     "--output",
@@ -334,8 +410,8 @@ def simgenotype(
     "-v",
     "--verbosity",
     type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]),
-    default="ERROR",
-    show_default="only errors",
+    default="INFO",
+    show_default=True,
     help="The level of verbosity desired",
 )
 def simphenotype(
@@ -344,14 +420,16 @@ def simphenotype(
     replications: int = 1,
     heritability: float = None,
     prevalence: float = None,
+    normalize: bool = True,
     region: str = None,
     samples: tuple[str] = tuple(),
     samples_file: Path = None,
     ids: tuple[str] = tuple(),
     ids_file: Path = None,
     chunk_size: int = None,
+    seed: int = None,
     output: Path = Path("-"),
-    verbosity: str = "ERROR",
+    verbosity: str = "INFO",
 ):
     """
     Haplotype-aware phenotype simulation. Create a set of simulated phenotypes from a
@@ -362,17 +440,11 @@ def simphenotype(
 
     Note: GENOTYPES must be the output from the transform subcommand.
     """
-    import logging
-
+    from .logging import getLogger
     from .sim_phenotype import simulate_pt
 
-    log = logging.getLogger("haptools simphenotype")
-    db_time = "|%(asctime)s" if verbosity == "DEBUG" else ""
-    logging.basicConfig(
-        format="[%(levelname)8s" + db_time + "] %(message)s (%(filename)s:%(lineno)s)",
-        level=verbosity,
-        datefmt="%H:%M:%S",
-    )
+    log = getLogger(name="simphenotype", level=verbosity)
+
     # handle samples
     if samples and samples_file:
         raise click.UsageError(
@@ -402,10 +474,12 @@ def simphenotype(
         replications,
         heritability,
         prevalence,
+        normalize,
         region,
         samples,
         ids,
         chunk_size,
+        seed,
         output,
         log,
     )
@@ -501,8 +575,8 @@ def simphenotype(
     "-v",
     "--verbosity",
     type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]),
-    default="ERROR",
-    show_default="only errors",
+    default="INFO",
+    show_default=True,
     help="The level of verbosity desired",
 )
 def transform(
@@ -525,17 +599,11 @@ def transform(
     GENOTYPES must be formatted as a VCF or PGEN and HAPLOTYPES must be formatted
     according to the .hap format spec
     """
-    import logging
-
+    from .logging import getLogger
     from .transform import transform_haps
 
-    log = logging.getLogger("haptools transform")
-    db_time = "|%(asctime)s" if verbosity == "DEBUG" else ""
-    logging.basicConfig(
-        format="[%(levelname)8s" + db_time + "] %(message)s (%(filename)s:%(lineno)s)",
-        level=verbosity,
-        datefmt="%H:%M:%S",
-    )
+    log = getLogger(name="transform", level=verbosity)
+
     # handle samples
     if samples and samples_file:
         raise click.UsageError(
@@ -670,8 +738,8 @@ def transform(
     "-v",
     "--verbosity",
     type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]),
-    default="ERROR",
-    show_default="only errors",
+    default="INFO",
+    show_default=True,
     help="The level of verbosity desired",
 )
 def ld(
@@ -702,17 +770,11 @@ def ld(
     If TARGET is a variant ID, the ID must appear in GENOTYPES. Otherwise, it must
     be present in the .hap file
     """
-    import logging
-
     from .ld import calc_ld
+    from .logging import getLogger
+
+    log = getLogger(name="ld", level=verbosity)
 
-    log = logging.getLogger("haptools ld")
-    db_time = "|%(asctime)s" if verbosity == "DEBUG" else ""
-    logging.basicConfig(
-        format="[%(levelname)8s" + db_time + "] %(message)s (%(filename)s:%(lineno)s)",
-        level=verbosity,
-        datefmt="%H:%M:%S",
-    )
     # handle samples
     if samples and samples_file:
         raise click.UsageError(
@@ -756,6 +818,7 @@ def ld(
     "--sort/--no-sort",
     is_flag=True,
     default=True,
+    show_default=True,
     help="Sorting of the file will not be performed",
 )
 @click.option(
@@ -770,8 +833,8 @@ def ld(
     "-v",
     "--verbosity",
     type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]),
-    default="ERROR",
-    show_default="only errors",
+    default="INFO",
+    show_default=True,
     help="The level of verbosity desired",
 )
 def index(
@@ -784,14 +847,10 @@ def index(
     Takes in an unsorted .hap file and outputs it as a .gz and a .tbi file
     """
 
-    import logging
     from .index import index_haps
+    from .logging import getLogger
 
-    log = logging.getLogger("haptools index")
-    logging.basicConfig(
-        format="[%(levelname)8s] %(message)s (%(filename)s:%(lineno)s)",
-        level=verbosity,
-    )
+    log = getLogger(name="index", level=verbosity)
 
     index_haps(haplotypes, sort, output, log)
 
